#
# Christmas Logic 
# Copyright (C) Jack Whitham 2009
#

from constants import *
import network, bresenham, dfl

class Network_Error(Exception):
    def __init__(self, terminal):
        self.terminal = terminal

    def Get_Terminal(self):
        return self.terminal

    def Get_Message(self):
        return "Base"

    def __str__(self):
        return self.Get_Message()

class Wired_Loop(Network_Error):
    def Get_Message(self):
        return ( "That move would create a loop within the wiring.\n"
            "Loops must include at least one 'Delay' component." )

class Wired_OR(Network_Error):
    def Get_Message(self):
        return ( "That would connect two outputs together.\n"
            "Outputs can never be linked to each other." )

class Combinatorial_Loop(Network_Error):
    def Get_Message(self):
        return ( "That move would create a loop.\n"
            "Loops must include at least one 'Delay' component." )

def Cell((x, y)):
    return (x >> CELL_SHIFT, y >> CELL_SHIFT)

class Network_Element:
    def __init__(self, n):
        assert isinstance(n, network.Network)
        self.drawable = False
        self.network = n
        self.network.Register(self)

    def Paint_Highlight(self, dc, wxp):
        pass

    def Paint(self, dc, wxp):
        pass

    def Refresh(self, wxwin):
        pass

    def Get_Cells(self):
        return []

    def Delete(self):
        #print 'delete', self
        self.network.Unregister(self)
        self.network = None

class Wire(Network_Element):
    def __init__(self, n, t1, t2):
        Network_Element.__init__(self, n)
        assert isinstance(t1, Terminal)
        assert isinstance(t2, Terminal)
        assert t1 != t2
        self.t1 = t1
        self.t2 = t2
        t1.wires.append(self)
        t2.wires.append(self)
        self.drawable = True

    def Refresh(self, wxwin):
        if self.drawable:
            (x1, y1) = self.t1.Get_Location()
            (x2, y2) = self.t2.Get_Location()
            if x2 < x1: (x1, x2) = (x2, x1)
            if y2 < y1: (y1, y2) = (y2, y1)
            x1 -= 5
            y1 -= 5
            x2 += 5
            y2 += 5
            wxwin.Update_Area((x1, y1), (x2 - x1, y2 - y1))

    def Paint(self, dc, wxp):
        (x1, y1) = self.t1.Get_Location()
        (x2, y2) = self.t2.Get_Location()
        dc.SetPen(wxp.PEN_REGULAR_LINE)
        dc.DrawLine(x1, y1, x2, y2)

    def Paint_Highlight(self, dc, wxp):
        (x1, y1) = self.t1.Get_Location()
        (x2, y2) = self.t2.Get_Location()
        dc.SetPen(wxp.PEN_HIGHLIGHT_LINE)
        dc.DrawLine(x1, y1, x2, y2)

    def Get_Cells(self):
        return bresenham.Line(Cell(self.t1.Get_Location()),
                    Cell(self.t2.Get_Location()))

    def Delete(self):
        Network_Element.Delete(self)
        for t in (self.t1, self.t2):
            assert isinstance(t, Terminal)
            assert self in t.wires
            for i in xrange(len(t.wires)):
                if t.wires[ i ] == self:
                    t.wires.pop(i)
                    break
            assert not (self in t.wires)
            if (isinstance(t.component, Split_Component)
            and (len(t.wires) == 0)):
                #print 'delete component'
                t.component.Delete()


class Terminal(Network_Element):
    def __init__(self, n, component, (dx, dy)):
        self.drawable = False
        Network_Element.__init__(self, n)
        assert isinstance(component, Base_Component)
        self.component = component
        self.dx = dx
        self.dy = dy
        self.wires = []
        self.updated_when = None
        self.drawable = True

    def Consider_Refresh(self, wxwin):
        s = self.Get_State() 
        if s != self.updated_when:
            self.updated_when = s
            self.Refresh(wxwin)

    def Get_Cells(self):
        yield Cell(self.Get_Location())

    def Get_Location(self):
        (x, y) = self.component.Get_Location()
        x += self.dx
        y += self.dy
        return (x, y)

    def Refresh(self, wxwin):
        if self.drawable:
            (x, y) = self.Get_Location()
            d = TERMINAL_RADIUS + HIGHLIGHT_INC
            d2 = d * 2
            wxwin.Update_Area((x - d, y - d), (d2, d2))

    def Paint(self, dc, wxp):
        (x, y) = self.Get_Location()
        dc.SetPen(wxp.PEN_REGULAR_TERMINAL)

        s = self.Get_State()
        if s == ST_UNDEFINED:
            dc.SetBrush(wxp.BR_UNDEFINED)
        elif s == ST_ONE:
            dc.SetBrush(wxp.BR_ONE)
        else:
            dc.SetBrush(wxp.BR_ZERO)

        dc.DrawCircle(x, y, TERMINAL_RADIUS)
        self.needs_update = True

    def Paint_Highlight(self, dc, wxp):
        (x, y) = self.Get_Location()
        dc.SetPen(wxp.PEN_HIGHLIGHT_TERMINAL)
        dc.SetBrush(wxp.BR_HIGHLIGHT_TERMINAL)
        dc.DrawCircle(x, y, TERMINAL_RADIUS + HIGHLIGHT_INC)
        self.needs_update = True

    def Get_State(self):
        return ST_UNDEFINED

    def Delete(self):
        Network_Element.Delete(self)
        for w in self.wires:
            w.Delete()

class In_Terminal(Terminal):
    def __init__(self, n, component, (dx, dy)):
        Terminal.__init__(self, n, component, (dx, dy))
        self.source = None

    def Get_State(self):
        if self.source == None:
            return ST_UNDEFINED

        return self.source.Get_State()

class Out_Terminal(Terminal):
    def __init__(self, n, component, (dx, dy)):
        Terminal.__init__(self, n, component, (dx, dy))
        self.state = ST_UNDEFINED

    def Get_State(self):
        return self.state

    def Set_State(self, x):
        self.state = x

class Base_Component(Network_Element):
    def __init__(self, n, (w, h)):
        self.drawable = False
        self.x = 0
        self.y = 0
        self.w = w
        self.h = h
        self.wh = w / 2
        self.hh = h / 2
        Network_Element.__init__(self, n)
        self.inputs = []
        self.outputs = []
        self.provides = []
        self.depth = 0
        self.state = ST_UNDEFINED

    def Get_Size(self):
        return (self.w, self.h)

    def Refresh(self, wxwin):
        if self.drawable:
            (x, y) = self.Get_Location()
            (w, h) = self.Get_Size()
            wxwin.Update_Area((x - (w / 2), y - (h / 2)), (w, h))

    def Get_Terminal(self):
        if len(self.outputs) == 1:
            assert len(self.inputs) == 0, "Which terminal?"
            return self.outputs[ 0 ]
        if len(self.inputs) == 1:
            return self.inputs[ 0 ]

        assert False, "Which terminal?"

    def Get_Cells(self):
        (cx1, cy1) = Cell((self.x - self.wh, self.y - self.hh))
        (cx2, cy2) = Cell((self.x + self.wh, self.y + self.hh))
        locs = []
        for cy in xrange(cy1, cy2 + 1, 1):
            for cx in xrange(cx1, cx2 + 1, 1):
                yield (cx, cy)

    def Get_State(self):
        return self.state

    def Reset_Depends(self):
        self.provides = []
        self.depth = -1

    def Update_Depends(self):
        for input_terminal in self.inputs:
            assert isinstance(input_terminal, In_Terminal)
            source_component = None
            wires = input_terminal.wires[:]
            seen_terminal = set()
            seen_terminal.add(input_terminal)

            while len(wires) != 0:
                wire = wires.pop()
                if wire.t1 in seen_terminal:
                    if wire.t2 in seen_terminal:
                        raise Wired_Loop(wire.t1)
                    else:
                        other_terminal = wire.t2
                elif wire.t2 in seen_terminal:
                    other_terminal = wire.t1
                else:
                    assert False, "Reached wire without reaching its terminal?"

                if isinstance(other_terminal, Out_Terminal):
                    if source_component != None:
                        raise Wired_OR(other_terminal)
                    source_component = other_terminal.component
                    input_terminal.source = other_terminal
                    assert source_component != None

                seen_terminal.add(other_terminal)
                for wire2 in other_terminal.wires:
                    if wire2 != wire:
                        wires.append(wire2)

            if ((source_component != None)
            and not (self in source_component.provides)):
                #print source_component, 'provides', self
                source_component.provides.append(self)

    def Get_Location(self):
        return (self.x, self.y)

    def Set_Location(self, (x, y)):
        # A move must involve deregistration and re-registration
        # of "self" and all wires connected to all terminals
        implicated = set()
        implicated.add(self)
        for tgroup in (self.inputs, self.outputs):
            for t in tgroup:
                for wire in t.wires:
                    implicated.add(wire)
                implicated.add(t)

        for t in implicated:
            t.network.Unregister(t)

        self.drawable = True
        
        # (x, y) is clipped to the maximum game area
        x = max(-WORLD_BOUNDARY, min(WORLD_BOUNDARY, x))
        y = max(-WORLD_BOUNDARY, min(WORLD_BOUNDARY, y))

        (self.x, self.y) = (x, y)

        for t in implicated:
            t.network.Register(t)


    def Paint(self, dc, wxp):
        pass

    def Paint_Highlight(self, dc, wxp):
        dc.SetPen(wxp.PEN_HIGHLIGHT_COMPONENT)
        dc.SetBrush(wxp.BR_HIGHLIGHT_COMPONENT)
        dc.DrawRectangle(self.x - self.wh, self.y - self.hh, self.w, self.h)

    def Set_State(self, s):
        self.state = s
        if len(self.outputs) != 0:
            self.outputs[ 0 ].Set_State(s)

    def Simulate(self):
        self.Set_State(ST_UNDEFINED)

    def Delete(self):
        Network_Element.Delete(self)
        for tgroup in (self.inputs, self.outputs):
            for t in tgroup:
                t.Delete()

class Special_Component(Base_Component):
    def __init__(self, n, (x, y)):
        Base_Component.__init__(self, n, (SPECIAL_SIZE, SPECIAL_SIZE))
        self.inputs = [ In_Terminal(n, self, (0, 0)) ]
        self.Set_Location((x, y))

    def Get_Terminal(self):
        return self.inputs[ 0 ]

class Split_Component(Special_Component):
    def Paint(self, dc, wxp):
        dc.SetPen(wxp.PEN_REGULAR)
        dc.SetBrush(wxp.BR_HOLLOW)
        dc.DrawRectangle(self.x + 2 - self.wh, 
                self.y + 2 - self.hh, self.w - 4, self.h - 4)

class Temporary_Component(Special_Component):
    pass

class Clocked_Component(Base_Component):
    def __init__(self, n, (w, h)):
        Base_Component.__init__(self, n, (w, h))

    def Tick(self):
        pass

class Not_Gate(Base_Component):
    def __init__(self, n):
        Base_Component.__init__(self, n, (100, 30))
        self.inputs = [ In_Terminal(n, self, (-self.wh, 0)) ]
        self.outputs = [ Out_Terminal(n, self, (self.wh, 0)) ]

    def Get_Symbol(self):
        return "NOT"
        
    def Paint(self, dc, wxp):
        (x, y) = (self.x - self.wh, self.y - self.hh)
        dc.SetPen(wxp.PEN_REGULAR)
        dc.SetBrush(wxp.BR_WHITE)
        dc.DrawLine(x, y + 15, x + 30, y + 15)
        dc.DrawRectangle(x + 30, y, 40, 30)
        dc.DrawLine(x + 70, y + 15, x + 100, y + 15)

        dc.SetFont(wxp.FONT_SMALL)
        dc.SetTextForeground((0, 0, 0))
        (w, h) = dc.GetTextExtent(self.Get_Symbol())
        dc.DrawText(self.Get_Symbol(), 
                    self.x - (w / 2), self.y - (h / 2))

    def Simulate(self):
        a = self.inputs[ 0 ].Get_State()
        if a == ST_UNDEFINED:
            self.Set_State(ST_UNDEFINED)
        elif a == ST_ONE:
            self.Set_State(ST_ZERO)
        else:
            self.Set_State(ST_ONE)

class Binary_Gate(Base_Component):
    def __init__(self, n):
        Base_Component.__init__(self, n, (100, 50))
        self.inputs = [ In_Terminal(n, self, (-self.wh, dy)) 
                for dy in [-20, 20] ]
        self.outputs = [ Out_Terminal(n, self, (self.wh, 0)) ]

    def Get_Symbol(self):
        return "base"
        
    def Paint(self, dc, wxp):
        (x, y) = (self.x - self.wh, self.y - self.hh)
        dc.SetPen(wxp.PEN_REGULAR)
        dc.SetBrush(wxp.BR_WHITE)
        dc.DrawLine(x, y + 5, x + 30, y + 5)
        dc.DrawLine(x, y + 45, x + 30, y + 45)
        dc.DrawRectangle(x + 30, y, 40, 50)
        dc.DrawLine(x + 70, y + 25, x + 100, y + 25)

        dc.SetFont(wxp.FONT_CL)
        dc.SetTextForeground((0, 0, 0))
        (w, h) = dc.GetTextExtent(self.Get_Symbol())
        dc.DrawText(self.Get_Symbol(), 
                    self.x - (w / 2), self.y - (h / 2))

class And_Gate(Binary_Gate):
    def Get_Symbol(self):
        return "&"

    def Simulate(self):
        a = self.inputs[ 0 ].Get_State()
        b = self.inputs[ 1 ].Get_State()
        if (ST_UNDEFINED in (a, b)):
            self.Set_State(ST_UNDEFINED)
        elif (a == b == ST_ONE):
            self.Set_State(ST_ONE)
        else:
            self.Set_State(ST_ZERO)

class Or_Gate(Binary_Gate):
    def Get_Symbol(self):
        return "+"

    def Simulate(self):
        a = self.inputs[ 0 ].Get_State()
        b = self.inputs[ 1 ].Get_State()
        if (ST_UNDEFINED in (a, b)):
            self.Set_State(ST_UNDEFINED)
        elif (ST_ONE in (a, b)):
            self.Set_State(ST_ONE)
        else:
            self.Set_State(ST_ZERO)

class Xor_Gate(Binary_Gate):
    def Get_Symbol(self):
        return "X"

    def Simulate(self):
        a = self.inputs[ 0 ].Get_State()
        b = self.inputs[ 1 ].Get_State()
        if (ST_UNDEFINED in (a, b)):
            self.Set_State(ST_UNDEFINED)
        elif (a != b):
            self.Set_State(ST_ONE)
        else:
            self.Set_State(ST_ZERO)

class Stimulator(Clocked_Component):
    def __init__(self, n, name, sequence):
        Clocked_Component.__init__(self, n, (80, 50))
        self.outputs = [ Out_Terminal(n, self, (self.wh, 0)) ]
        self.symbol = name
        self.pointer = 0
        self.sequence = []
        for s in sequence:
            if s == "1":
                self.sequence.append(ST_ONE)
            elif s == "0":
                self.sequence.append(ST_ZERO)
            else:
                assert False
        self.next_state = self.sequence[ 0 ]

    def Get_Symbol(self):
        return self.symbol

    def Tick(self):
        self.next_state = self.sequence[ self.pointer % len(self.sequence) ]

    def Paint(self, dc, wxp):
        (x, y) = (self.x - self.wh, self.y - self.hh)
        dc.SetPen(wxp.PEN_REGULAR)
        dc.SetBrush(wxp.BR_WHITE)
        dc.DrawLine(x + 60, y + 25, x + 80, y + 25)
        dc.DrawRectangle(x, y, 60, 50)

        dc.SetFont(wxp.FONT_SMALL)
        dc.SetTextForeground((0, 0, 0))
        for (y, text) in [
                        (-10, self.symbol),
                        (10, "input"),
                        ]:
            (w, h) = dc.GetTextExtent(text)
            dc.DrawText(text, x + 30 - (w / 2), self.y + y - (h / 2))
        
    def Simulate(self):
        self.Set_State(self.next_state)

class Stimulator_Simple(Stimulator):
    def __init__(self, n, initial_state):
        if initial_state == ST_CLOCK:
            sequence = "10"
            name = "flashing"
        elif initial_state == ST_ONE:
            sequence = "1"
            name = "power"
        else:
            assert initial_state == ST_ZERO
            sequence = "0"
            name = "off"

        Stimulator.__init__(self, n, name, sequence)

class Light_Control(Base_Component):
    def __init__(self, n, level_string):
        Base_Component.__init__(self, n, (80, 50))
        self.inputs = [ In_Terminal(n, self, (-self.wh, 0)) ]
        self.level_string = level_string

    def Paint(self, dc, wxp):
        (x, y) = (self.x - self.wh, self.y - self.hh)

        dc.SetPen(wxp.PEN_REGULAR)
        dc.DrawLine(x, y + 25, x + 20, y + 25)
        dc.SetBrush(wxp.Get_Brush(self.level_string.colour))
        dc.DrawRectangle(x + 20, y, 60, 50)

        dc.SetPen(wxp.PEN_TRANSPARENT)
        dc.SetBrush(wxp.BR_WHITE)
        dc.DrawRectangle(x + 25, y + 5, 60 - 10, 50 - 10)

        dc.SetFont(wxp.FONT_SMALL)
        dc.SetTextForeground((0, 0, 0))
        for (y, text) in [
                        (-10, self.level_string.name),
                        (10, "lights"),
                        ]:
            (w, h) = dc.GetTextExtent(text)
            dc.DrawText(text, x + 50 - (w / 2), self.y + y - (h / 2))
        
    def Simulate(self):
        self.Set_State(self.inputs[ 0 ].Get_State())

    def Set_State(self, s):
        Base_Component.Set_State(self, s)
        self.network.Set_Light_State(self.level_string.name, s)

class D_Flip_Flop(Clocked_Component):
    def __init__(self, n, ce = True):
        Clocked_Component.__init__(self, n, (100, 50))
        self.inputs = [ In_Terminal(n, self, (-self.wh, -20)) ]
        self.outputs = [ Out_Terminal(n, self, (self.wh, 0)) ]
        self.dce = [ (-18, "d") ]
        self.next_state = ST_UNDEFINED
        if ce:
            self.inputs.append(In_Terminal(n, self, (-self.wh, 20)))
            self.dce.append((18, "ce"))

    def Paint(self, dc, wxp):
        (x, y) = (self.x - self.wh, self.y - self.hh)
        dc.SetPen(wxp.PEN_REGULAR)
        dc.SetBrush(wxp.BR_WHITE)
        dc.DrawLine(x + 70, y + 25, x + 100, y + 25)
        dc.DrawRectangle(x + 30, y, 40, 50)

        text = "D"
        dc.SetFont(wxp.FONT_CL)
        dc.SetTextForeground((0, 0, 0))
        (w, h) = dc.GetTextExtent(text)
        dc.DrawText(text, self.x + 5 - (w / 2), self.y - (h / 2))

        dc.SetFont(wxp.FONT_SMALL)
        for (y, text) in self.dce:
            (w, h) = dc.GetTextExtent(text)
            y += self.y
            dc.DrawText(text, x + 30, y - (h / 2))
            dc.DrawLine(x, y, x + 30, y)
        
    def Simulate(self):
        d = self.inputs[ 0 ].Get_State()
        ce = ST_ONE
        if len(self.inputs) > 1:
            ce = self.inputs[ 1 ].Get_State()
            
        if ((ce == ST_UNDEFINED)
        or (ce == ST_ZERO)):
            d = self.Get_State() # No update

        elif ((d == ST_UNDEFINED)
        and (len(self.inputs[ 0 ].wires) != 0)):
            d = ST_ZERO # self-starting

        self.next_state = d

    def Tick(self):
        self.Set_State(self.next_state)

class D_Flip_Flop_NCE(D_Flip_Flop):
    def __init__(self, n):
        D_Flip_Flop.__init__(self, n, False)

class Guide_Component(Base_Component):
    def __init__(self, n, guide_data, level_vars):
        Base_Component.__init__(self, n, (1, 1))
        self.level_vars = level_vars
        self.guide_data = guide_data
        self.level_vars[ "n" ] = self.network
        self.Reset()

    def Reset(self):
        self.old_ack_counter = 0
        for g in self.guide_data:
            g.received_ack = False

    def Update(self, ack_counter):
        import baselevel 
        env = baselevel.__dict__

        contents = None
        point_at = None
        requires_ack = False
        post_ack = False
        self.level_vars[ "s" ] = self.network.Get_Selected_Item()

        if ack_counter != self.old_ack_counter:
            self.old_ack_counter = ack_counter
            post_ack = True

        for g in self.guide_data:
            if g.requires_ack:
                activate = True
                if (not g.received_ack) and post_ack:
                    g.received_ack = True
                    post_ack = False
            else:
                activate = eval(g.trigger_expression, env, self.level_vars)

            if g.received_ack:
                continue

            if activate:
                contents = g.text
                p = eval(g.point_at_expression, 
                                    env, self.level_vars)
                if p != None:
                    point_at = p.Get_Location()
                else:
                    point_at = None

                if g.requires_ack:
                    requires_ack = True
                    break

        return {
            "requires_ack" : requires_ack,
            "point_at" : point_at,
            "contents" : contents,
            "location" : self.Get_Location(),
            }

    def Refresh(self, wxwin):
        pass

    def Paint_Highlight(self, dc, wxp):
        pass

    def Paint(self, dc, wxp):
        pass
        
class Splash_Component(Clocked_Component):
    def __init__(self, n, ack_text, target_list):
        Clocked_Component.__init__(self, n, (644, 436))

        self.program = []
        self.outputs = []

        dy = 100
        y = - ((len(target_list) - 1) * dy / 2)
        for i in xrange(len(target_list)):
            self.outputs.append(Out_Terminal(n, self, (self.wh, y)))
            y += dy
            Wire(self.network, self.outputs[ i ], 
                    target_list[ i ].Get_Terminal())

            gs = target_list[ i ].level_string.goal_sequence[:]
            self.program.append(gs)
            if len(self.program) > 1:
                assert len(self.program[ 0 ]) == len(self.program[ -1 ])

        self.ack_text = ack_text
        self.pointer = 0

    def Tick(self):
        self.pointer += 1

    def Update(self, ack_counter):
        return {
                "location" : self.Get_Location(),
                "contents" : "splash",
                "ack_text" : self.ack_text,
                "requires_ack" : True,
            }

    def Paint(self, dc, wxp):
        pass

    def Paint_Highlight(self, dc, wxp):
        pass

    def Simulate(self):
        self.Set_State(ST_ZERO)

        k = len(self.program[ 0 ])
        j = (self.pointer + k - 3) % k
        for i in xrange(len(self.outputs)):
            self.outputs[ i ].Set_State(self.program[ i ][ j ])

class End_Component(Splash_Component):
    def Update(self, ack_counter):
        x = Splash_Component.Update(self, ack_counter)
        x[ "contents" ] = "end"
        return x

